==> pyenphase-1.1.3-py3-none-any.whl: pyenphase/envoy.py <==
import enum
import logging
from typing import Any

import httpx
import orjson
from awesomeversion import AwesomeVersion
from envoy_utils.envoy_utils import EnvoyUtils
from tenacity import retry, retry_if_exception_type, wait_random_exponential

from .auth import EnvoyAuth, EnvoyLegacyAuth, EnvoyTokenAuth
from .const import (
    LOCAL_TIMEOUT,
    URL_DRY_CONTACT_SETTINGS,
    URL_DRY_CONTACT_STATUS,
    URL_ENCHARGE_BATTERY,
    URL_ENSEMBLE_INVENTORY,
    URL_PRODUCTION,
    URL_PRODUCTION_INVERTERS,
    URL_PRODUCTION_JSON,
    URL_PRODUCTION_V1,
)
from .exceptions import (
    ENDPOINT_PROBE_EXCEPTIONS,
    EnvoyAuthenticationRequired,
    EnvoyProbeFailed,
)
from .firmware import EnvoyFirmware
from .models.dry_contacts import EnvoyDryContactSettings, EnvoyDryContactStatus
from .models.encharge import EnvoyEncharge, EnvoyEnchargePower
from .models.enpower import EnvoyEnpower
from .models.envoy import EnvoyData
from .models.inverter import EnvoyInverter
from .models.system_consumption import EnvoySystemConsumption
from .models.system_production import EnvoySystemProduction
from .ssl import NO_VERIFY_SSL_CONTEXT

_LOGGER = logging.getLogger(__name__)


LEGACY_ENVOY_VERSION = AwesomeVersion("3.9.0")
ENSEMBLE_MIN_VERSION = AwesomeVersion("5.0.0")
AUTH_TOKEN_MIN_VERSION = AwesomeVersion("7.0.0")
DEFAULT_HEADERS = {
    "Accept": "application/json",
}


class SupportedFeatures(enum.IntFlag):
    INVERTERS = 1
    METERING = 2
    TOTAL_CONSUMPTION = 4
    NET_CONSUMPTION = 8
    ENCHARGE = 16
    ENPOWER = 32


class Envoy:
    """Class for communicating with an envoy."""

    def __init__(
        self,
        host: str,
        client: httpx.AsyncClient | None = None,
        timeout: float | httpx.Timeout | None = None,
    ) -> None:
        """Initialize the Envoy class."""
        # We use our own httpx client session so we can disable SSL verification (Envoys use self-signed SSL certs)
        self._timeout = timeout or LOCAL_TIMEOUT
        self._client = client or httpx.AsyncClient(
            verify=NO_VERIFY_SSL_CONTEXT
        )  # nosec
        self.auth: EnvoyAuth | None = None
        self._host = host
        self._firmware = EnvoyFirmware(self._client, self._host)
        self._supported_features: SupportedFeatures = SupportedFeatures(0)
        self._production_endpoint: str | None = None
        self._consumption_endpoint: str | None = None
        self.data: EnvoyData | None = None

    async def setup(self) -> None:
        """Obtain the firmware version for later Envoy authentication."""
        await self._firmware.setup()

    async def authenticate(
        self,
        username: str | None = None,
        password: str | None = None,
        token: str | None = None,
    ) -> None:
        """Authenticate to the Envoy based on firmware version."""
        if self._firmware.version < AUTH_TOKEN_MIN_VERSION:
            # Envoy firmware using old envoy/installer authentication
            _LOGGER.debug(
                "Authenticating to Envoy using envoy/installer authentication"
            )
            full_serial = self._firmware.serial
            if not username or username == "installer":
                username = "installer"
                password = EnvoyUtils.get_password(full_serial, username)
            elif username == "envoy" and not password:
                # The default password for the envoy user is the first 6 digits of the serial number
                assert full_serial is not None, "Serial must be set"  # nosec
                password = full_serial[:6]

            if username and password:
                self.auth = EnvoyLegacyAuth(self.host, username, password)

        else:
            # Envoy firmware using new token authentication
            _LOGGER.debug("Authenticating to Envoy using token authentication")
            if token or (username and password):
                # Always pass all the data to the token auth class, even if some of it is None
                # so that we can refresh the token if needed
                self.auth = EnvoyTokenAuth(
                    self.host,
                    cloud_username=username,
                    cloud_password=password,
                    envoy_serial=self._firmware.serial,
                    token=token,
                )

        if not self.auth:
            _LOGGER.error(
                "You must include username/password or token to authenticate to the Envoy."
            )
            raise EnvoyAuthenticationRequired("Could not setup authentication object.")

        await self.auth.setup(self._client)

    @retry(
        retry=retry_if_exception_type(
            (httpx.NetworkError, httpx.TimeoutException, httpx.RemoteProtocolError)
        ),
        wait=wait_random_exponential(multiplier=2, max=3),
    )
    async def probe_request(self, endpoint: str) -> httpx.Response:
        """Make a probe request.

        Probe requests are not retried on bad JSON responses.
        """
        return await self._request(endpoint)

    @retry(
        retry=retry_if_exception_type(
            (
                httpx.NetworkError,
                httpx.TimeoutException,
                httpx.RemoteProtocolError,
                orjson.JSONDecodeError,
            )
        ),
        wait=wait_random_exponential(multiplier=2, max=3),
    )
    async def request(self, endpoint: str) -> Any:
        """Make a request to the Envoy.

        Request retries on bad JSON responses which the Envoy sometimes returns.
        """
        return await self._request(endpoint)

    async def _request(self, endpoint: str) -> Any:
        """Make a request to the Envoy."""
        if self.auth is None:
            raise EnvoyAuthenticationRequired(
                "You must authenticate to the Envoy before making requests."
            )

        url = self.auth.get_endpoint_url(endpoint)
        _LOGGER.debug("Requesting %s with timeout %s", url, self._timeout)
        response = await self._client.get(
            url,
            headers={**DEFAULT_HEADERS, **self.auth.headers},
            cookies=self.auth.cookies,
            follow_redirects=True,
            auth=self.auth.auth,
            timeout=self._timeout,
        )
        try:
            return orjson.loads(response.text)
        except orjson.JSONDecodeError as e:
            _LOGGER.debug(
                "Unable to decode response from Envoy endpoint %s: %s", url, e
            )
            raise

    @property
    def host(self) -> str:
        """Return the Envoy host."""
        return self._host

    @property
    def firmware(self) -> AwesomeVersion:
        """Return the Envoy firmware version."""
        return self._firmware.version

    @property
    def part_number(self) -> str | None:
        """Return the Envoy part number."""
        return self._firmware.part_number

    @property
    def serial_number(self) -> str | None:
        """Return the Envoy serial number."""
        return self._firmware.serial

    async def probe(self) -> None:
        """Probe for model and supported features."""
        for possible_endpoint in (URL_PRODUCTION, URL_PRODUCTION_JSON):
            try:
                production_json: dict[str, Any] = await self.probe_request(
                    possible_endpoint
                )
            except ENDPOINT_PROBE_EXCEPTIONS as e:
                _LOGGER.debug(
                    "Production endpoint not found at %s: %s", possible_endpoint, e
                )
                continue
            else:
                production: list[
                    dict[str, str | float | int]
                ] | None = production_json.get("production")
                if production:
                    for type_ in production:
                        if type_["type"] == "eim" and type_["activeCount"]:
                            self._supported_features |= SupportedFeatures.METERING
                            self._production_endpoint = possible_endpoint
                            break

                consumption: list[
                    dict[str, str | float | int]
                ] | None = production_json.get("consumption")
                if consumption:
                    for meter in consumption:
                        meter_type = meter["measurementType"]
                        if meter_type == "total-consumption":
                            self._supported_features |= (
                                SupportedFeatures.TOTAL_CONSUMPTION
                            )
                            if not self._consumption_endpoint:
                                self._consumption_endpoint = possible_endpoint
                        elif meter_type == "net-consumption":
                            self._supported_features |= (
                                SupportedFeatures.NET_CONSUMPTION
                            )
                            if not self._consumption_endpoint:
                                self._consumption_endpoint = possible_endpoint

                break

        if not self._production_endpoint:
            try:
                await self.probe_request(URL_PRODUCTION_V1)
            except ENDPOINT_PROBE_EXCEPTIONS as e:
                _LOGGER.debug(
                    "Production endpoint not found at %s: %s", URL_PRODUCTION_V1, e
                )
            else:
                self._production_endpoint = URL_PRODUCTION_V1

        if not self._production_endpoint:
            raise EnvoyProbeFailed("Unable to determine production endpoint")

        try:
            await self.probe_request(URL_PRODUCTION_INVERTERS)
        except ENDPOINT_PROBE_EXCEPTIONS as e:
            _LOGGER.debug("Inverters endpoint not found: %s", e)
        else:
            self._supported_features |= SupportedFeatures.INVERTERS

        # Check for various Ensemble support
        if self._firmware.version >= ENSEMBLE_MIN_VERSION:
            # The Ensemble Inventory endpoint will tell us if we have Enpower or Encharge support
            try:
                result = await self.probe_request(URL_ENSEMBLE_INVENTORY)
            except ENDPOINT_PROBE_EXCEPTIONS as e:
                _LOGGER.debug("Ensemble Inventory endpoint not found: %s", e)
            else:
                if not result or "error" in result:
                    # Newer firmware with no Ensemble devices returns an empty list
                    _LOGGER.debug("No Ensemble devices found")
                    return

                for item in result:
                    if item["type"] == "ENPOWER":
                        self._supported_features |= SupportedFeatures.ENPOWER
                    if item["type"] == "ENCHARGE":
                        self._supported_features |= SupportedFeatures.ENCHARGE
        else:
            _LOGGER.debug("Firmware too old for Ensemble support")

    async def update(self) -> EnvoyData:
        """Update data."""
        if not self._production_endpoint:
            await self.probe()

        production_endpoint = self._production_endpoint
        consumption_endpoint = self._consumption_endpoint
        supported_features = self._supported_features

        production_data = await self.request(production_endpoint)
        raw = {"production": production_data}
        if consumption_endpoint == production_endpoint:
            consumption_data = production_data
        elif consumption_endpoint:
            consumption_data = await self.request(consumption_endpoint)
            raw["consumption"] = consumption_data

        inverters: dict[str, EnvoyInverter] = {}

        if production_endpoint in (URL_PRODUCTION, URL_PRODUCTION_JSON):
            system_production = EnvoySystemProduction.from_production(production_data)
        else:
            # Production endpoint is URL_PRODUCTION_V1
            system_production = EnvoySystemProduction.from_v1_api(production_data)

        system_consumption: EnvoySystemConsumption | None = None
        if consumption_endpoint:
            system_consumption = EnvoySystemConsumption.from_production(
                consumption_data
            )

        if supported_features & SupportedFeatures.INVERTERS:
            inverters_data: list[dict[str, Any]] = await self.request(
                URL_PRODUCTION_INVERTERS
            )
            inverters = {
                inverter["serialNumber"]: EnvoyInverter.from_v1_api(inverter)
                for inverter in inverters_data
            }
            raw["inverters"] = inverters_data

        # Update Enpower and Encharge data if supported
        encharge_inventory: dict[str, EnvoyEncharge] | None = None
        encharge_power: dict[str, EnvoyEnchargePower] | None = None
        enpower: EnvoyEnpower | None = None
        dry_contact_settings: dict[str, EnvoyDryContactSettings] = {}
        dry_contact_status: dict[str, EnvoyDryContactStatus] = {}

        if (
            supported_features & SupportedFeatures.ENCHARGE
            or supported_features & SupportedFeatures.ENPOWER
        ):
            ensemble_inventory_data: list[dict[str, Any]] = await self.request(
                URL_ENSEMBLE_INVENTORY
            )
            raw["ensemble"] = ensemble_inventory_data

            if supported_features & SupportedFeatures.ENCHARGE:
                encharge_power_data: dict[str, Any] = await self.request(
                    URL_ENCHARGE_BATTERY
                )
                raw["encharge_power"] = encharge_power_data
                power: dict[str, Any] = {
                    power["serial_num"]: power
                    for power in encharge_power_data["devices:"]
                }
                inventory: dict[str, Any] = {}
                for item in ensemble_inventory_data:
                    if item["type"] != "ENCHARGE":
                        continue
                    inventory = {
                        device["serial_num"]: device for device in item["devices"]
                    }

                encharge_inventory = {
                    serial: EnvoyEncharge.from_api(inventory[serial])
                    for serial in inventory
                }
                encharge_power = {
                    serial: EnvoyEnchargePower.from_api(power[serial])
                    for serial in power
                }
            if supported_features & SupportedFeatures.ENPOWER:
                # Update Enpower data
                for item in ensemble_inventory_data:
                    if item["type"] != "ENPOWER":
                        continue
                    enpower_data = item["devices"][0]
                raw["enpower"] = enpower_data
                enpower = EnvoyEnpower.from_api(enpower_data)

                # Update dry contact data
                dry_contact_status_data = await self.request(URL_DRY_CONTACT_STATUS)
                dry_contact_settings_data = await self.request(URL_DRY_CONTACT_SETTINGS)
                raw["dry_contact_status"] = dry_contact_status_data
                raw["dry_contact_settings"] = dry_contact_settings_data

                dry_contact_status = {
                    relay["id"]: EnvoyDryContactStatus.from_api(relay)
                    for relay in dry_contact_status_data["dry_contacts"]
                }
                dry_contact_settings = {
                    relay["id"]: EnvoyDryContactSettings.from_api(relay)
                    for relay in dry_contact_settings_data["dry_contacts"]
                }

        data = EnvoyData(
            system_production=system_production,
            system_consumption=system_consumption,
            inverters=inverters,
            encharge_inventory=encharge_inventory,
            encharge_power=encharge_power,
            enpower=enpower,
            dry_contact_status=dry_contact_status,
            dry_contact_settings=dry_contact_settings,
            # Raw data is exposed so we can __eq__ the data to see if
            # anything has changed and consumers of the library can
            # avoid dispatching data if nothing has changed.
            raw=raw,
        )
        self.data = data
        return data

==> pyenphase-1.3.0-py3-none-any.whl: pyenphase/envoy.py <==
import logging
from collections.abc import Callable

import httpx
import orjson
from awesomeversion import AwesomeVersion
from envoy_utils.envoy_utils import EnvoyUtils
from tenacity import retry, retry_if_exception_type, wait_random_exponential

from .auth import EnvoyAuth, EnvoyLegacyAuth, EnvoyTokenAuth
from .const import AUTH_TOKEN_MIN_VERSION, LOCAL_TIMEOUT, SupportedFeatures
from .exceptions import EnvoyAuthenticationRequired, EnvoyProbeFailed
from .firmware import EnvoyFirmware
from .models.envoy import EnvoyData
from .ssl import NO_VERIFY_SSL_CONTEXT
from .updaters.api_v1_production import EnvoyApiV1ProductionUpdater
from .updaters.api_v1_production_inverters import EnvoyApiV1ProductionInvertersUpdater
from .updaters.base import EnvoyUpdater
from .updaters.ensemble import EnvoyEnembleUpdater
from .updaters.production import EnvoyProductionJsonUpdater, EnvoyProductionUpdater

_LOGGER = logging.getLogger(__name__)


DEFAULT_HEADERS = {
    "Accept": "application/json",
}

UPDATERS: list[type["EnvoyUpdater"]] = [
    EnvoyProductionUpdater,
    EnvoyProductionJsonUpdater,
    EnvoyApiV1ProductionUpdater,
    EnvoyApiV1ProductionInvertersUpdater,
    EnvoyEnembleUpdater,
]


def register_updater(updater: type[EnvoyUpdater]) -> Callable[[], None]:
    """Register an updater."""
    UPDATERS.append(updater)

    def _remove_updater() -> None:
        UPDATERS.remove(updater)

    return _remove_updater


def get_updaters() -> list[type[EnvoyUpdater]]:
    return UPDATERS


class Envoy:
    """Class for communicating with an envoy."""

    def __init__(
        self,
        host: str,
        client: httpx.AsyncClient | None = None,
        timeout: float | httpx.Timeout | None = None,
    ) -> None:
        """Initialize the Envoy class."""
        # We use our own httpx client session so we can disable SSL verification (Envoys use self-signed SSL certs)
        self._timeout = timeout or LOCAL_TIMEOUT
        self._client = client or httpx.AsyncClient(
            verify=NO_VERIFY_SSL_CONTEXT
        )  # nosec
        self.auth: EnvoyAuth | None = None
        self._host = host
        self._firmware = EnvoyFirmware(self._client, self._host)
        self._supported_features: SupportedFeatures | None = None
        self._updaters: list[EnvoyUpdater] = []
        self.data: EnvoyData | None = None

    async def setup(self) -> None:
        """Obtain the firmware version for later Envoy authentication."""
        await self._firmware.setup()

    async def authenticate(
        self,
        username: str | None = None,
        password: str | None = None,
        token: str | None = None,
    ) -> None:
        """Authenticate to the Envoy based on firmware version."""
        if self._firmware.version < AUTH_TOKEN_MIN_VERSION:
            # Envoy firmware using old envoy/installer authentication
            _LOGGER.debug(
                "Authenticating to Envoy using envoy/installer authentication"
            )
            full_serial = self._firmware.serial
            if not username or username == "installer":
                username = "installer"
                password = EnvoyUtils.get_password(full_serial, username)
            elif username == "envoy" and not password:
                # The default password for the envoy user is the first 6 digits of the serial number
                assert full_serial is not None, "Serial must be set"  # nosec
                password = full_serial[:6]

            if username and password:
                self.auth = EnvoyLegacyAuth(self.host, username, password)

        else:
            # Envoy firmware using new token authentication
            _LOGGER.debug("Authenticating to Envoy using token authentication")
            if token or (username and password):
                # Always pass all the data to the token auth class, even if some of it is None
                # so that we can refresh the token if needed
                self.auth = EnvoyTokenAuth(
                    self.host,
                    cloud_username=username,
                    cloud_password=password,
                    envoy_serial=self._firmware.serial,
                    token=token,
                )

        if not self.auth:
            _LOGGER.error(
                "You must include username/password or token to authenticate to the Envoy."
            )
            raise EnvoyAuthenticationRequired("Could not setup authentication object.")

        await self.auth.setup(self._client)

    @retry(
        retry=retry_if_exception_type(
            (httpx.NetworkError, httpx.TimeoutException, httpx.RemoteProtocolError)
        ),
        wait=wait_random_exponential(multiplier=2, max=3),
    )
    async def probe_request(self, endpoint: str) -> httpx.Response:
        """Make a probe request.

        Probe requests are not retried on bad JSON responses.
        """
        return await self._request(endpoint)

    @retry(
        retry=retry_if_exception_type(
            (
                httpx.NetworkError,
                httpx.TimeoutException,
                httpx.RemoteProtocolError,
                orjson.JSONDecodeError,
            )
        ),
        wait=wait_random_exponential(multiplier=2, max=3),
    )
    async def request(self, endpoint: str) -> httpx.Response:
        """Make a request to the Envoy.

        Request retries on bad JSON responses which the Envoy sometimes returns.
        """
        return await self._request(endpoint)

    async def _request(self, endpoint: str) -> httpx.Response:
        """Make a request to the Envoy."""
        if self.auth is None:
            raise EnvoyAuthenticationRequired(
                "You must authenticate to the Envoy before making requests."
            )

        url = self.auth.get_endpoint_url(endpoint)
        _LOGGER.debug("Requesting %s with timeout %s", url, self._timeout)
        return await self._client.get(
            url,
            headers={**DEFAULT_HEADERS, **self.auth.headers},
            cookies=self.auth.cookies,
            follow_redirects=True,
            auth=self.auth.auth,
            timeout=self._timeout,
        )

    @property
    def host(self) -> str:
        """Return the Envoy host."""
        return self._host

    @property
    def firmware(self) -> AwesomeVersion:
        """Return the Envoy firmware version."""
        return self._firmware.version

    @property
    def part_number(self) -> str | None:
        """Return the Envoy part number."""
        return self._firmware.part_number

    @property
    def serial_number(self) -> str | None:
        """Return the Envoy serial number."""
        return self._firmware.serial

    async def probe(self) -> None:
        """Probe for model and supported features."""
        supported_features = SupportedFeatures(0)
        updaters: list[EnvoyUpdater] = []
        version = self._firmware.version
        for updater in get_updaters():
            klass = updater(version, self.probe_request, self.request)
            if updater_features := await klass.probe(supported_features):
                supported_features |= updater_features
                updaters.append(klass)

        if not supported_features & SupportedFeatures.PRODUCTION:
            raise EnvoyProbeFailed("Unable to determine production endpoint")

        self._updaters = updaters
        self._supported_features = supported_features

    async def update(self) -> EnvoyData:
        """Update data."""
        if not self._supported_features:
            await self.probe()

        data = EnvoyData()
        for updater in self._updaters:
            await updater.update(data)

        self.data = data
        return data
